#include "Z80.h"
#include "computer.h"
#include "emulator.h"
#include "cargador.h"
#include <stdio.h>
#include <string.h>

void uncompress_z80(FILE *fichero,int length,unsigned char *memo) {

  unsigned char byte_loaded,EDfound,counter;
  int position;

  counter=0;
  byte_loaded=0;
  EDfound=0;
  position=0;
   
  do {
    if(counter) {
      memo[position++]=byte_loaded;
      counter--;
      continue;
    } else
      fread(&byte_loaded,1,1,fichero);
    
    if(EDfound==2) { // we have two EDs
      counter=byte_loaded;
      fread(&byte_loaded,1,1,fichero);
      EDfound=0;
      continue;
    }
    
    if(byte_loaded==0xED) {
      EDfound++;	
    } else {
      if(EDfound==1) { // we found ED xx. We write ED and xx
	memo[position++]=0xED;
	EDfound=0;
      }
      memo[position++]=byte_loaded;
    }
  } while(position<length);

}

int save_z80(char *filename) {

  FILE *fichero;
  unsigned char value,bucle;

  fichero=fopen(filename,"r");
  if(fichero!=NULL) {
    fclose(fichero);
    return -1; // file already exists
  }

  fichero=fopen(filename,"w");

  if(fichero==NULL)
    return -2; // can't create file

  fprintf(fichero,"%c%c%c%c%c%c",procesador.AF.B.h,procesador.AF.B.l,procesador.BC.B.l,procesador.BC.B.h,procesador.HL.B.l,procesador.HL.B.h); // AF, BC and HL

  if(ordenador.mode128k==0) // 48K
    fprintf(fichero,"%c%c",procesador.PC.B.l,procesador.PC.B.h); // PC
  else
    fprintf(fichero,"%c%c",0,0); // 128K

  fprintf(fichero,"%c%c",procesador.SP.B.l,procesador.SP.B.h); // SP

  fprintf(fichero,"%c%c%c",procesador.I,procesador.R,(((procesador.R>>7)&0x01)|((ordenador.border<<1)&0x0E))); // I, R and border color

  fprintf(fichero,"%c%c%c%c%c%c%c%c%c%c%c%c%c%c",procesador.DE.B.l,procesador.DE.B.h,procesador.BC1.B.l,procesador.BC1.B.h,procesador.DE1.B.l,procesador.DE1.B.h,procesador.HL1.B.l,procesador.HL1.B.h,procesador.AF1.B.h,procesador.AF1.B.l,procesador.IY.B.l,procesador.IY.B.h,procesador.IX.B.l,procesador.IX.B.h);

  if (procesador.IFF&IFF_1)
    value=1;
  else
    value=0;
  fprintf(fichero,"%c",value);
  if (procesador.IFF&IFF_2)
    value=1;
  else
    value=0;
  fprintf(fichero,"%c",value);
  if (procesador.IFF&IFF_IM1)
    value=1;
  else
    if(procesador.IFF&IFF_IM2)
      value=2;
    else
      value=0;
  if(ordenador.issue==2)
    value|=4;
  fprintf(fichero,"%c",value);

  if(ordenador.mode128k==0) { // 48K
    fwrite((ordenador.memoria+147456),16384,1,fichero); // video memory
    fwrite((ordenador.memoria+98304),32768,1,fichero); // memory pages 2 & 3
    fclose(fichero);
    return 0;
  }

  // 128K

  fprintf(fichero,"%c%c",23,0); // Z80 file v2.01
  fprintf(fichero,"%c%c",procesador.PC.B.l,procesador.PC.B.h); // PC
  fprintf(fichero,"%c",3); // hardware mode=3
  fprintf(fichero,"%c",ordenador.mport1); // content of 0x7FFD latch
  fprintf(fichero,"%c%c",0,0); // no If1, no emulation of any kind
  fprintf(fichero,"%c",ordenador.ay_latch); // last selected AY register
  for(bucle=0;bucle<16;bucle++)
    fprintf(fichero,"%c",ordenador.ay_registers[bucle]); // AY registers
  for(bucle=0;bucle<8;bucle++) {
    fprintf(fichero,"%c%c",0,0x40); // length=16384
    fprintf(fichero,"%c",bucle+3); // page number
    fwrite(ordenador.memoria+(16384*bucle)+65536,16384,1,fichero); // store page
  }
  fclose(fichero);
  return 0;
}

int load_z80(char *filename) {

  struct z80snapshot snap;
  unsigned char tempo[30],tempo2[56],memo[49152],type,compressed,page,byte_read[3];
  FILE *fichero;
  int longitud=0,longitud2,bucle;

  for(bucle=0;bucle<16;bucle++)
    snap.ay_regs[bucle]=0;

  fichero=fopen(filename,"r");
  if(fichero==NULL)
    return -1; // error


  fread(tempo,1,30,fichero);

  if((tempo[6]==0)&&(tempo[7]==0)) { // extended Z80

    fread(tempo2,1,2,fichero); // read the length of the extension
 
    longitud=((int)tempo2[0])+256*((int)tempo2[1]);
    if(longitud>54) {
      fclose(fichero);
      return -3; // not a supported Z80 file
    }

    fread(tempo2+2,1,longitud,fichero);

    type=1; // new type

    if(longitud==23) // z80 ver 2.01
      switch(tempo2[4]) {
      case 0:
      case 1:
	snap.type=0; // 48K
	break;
      case 3:
      case 4:
	snap.type=1; // 128K
	break;
      default:
	fclose(fichero);
	return -3; // not a supported Z80 file
	break;
      }
    else // z80 ver 3.0x
      switch(tempo2[4]) {
      case 0:
      case 1:
      case 3:
	snap.type=0; // 48K
	break;
      case 4:
      case 5:
      case 6:
	snap.type=1; // 128K
	break;
      default:
	fclose(fichero);
	return -3; // not a supported Z80 file
	break;
      }      
  } else {
    type=0; // old type
    snap.type=0; // 48k
  }

  if(tempo[29]&0x04)
    snap.issue=2;
  else
    snap.issue=3;

  snap.A=tempo[0];
  snap.F=tempo[1];
  snap.C=tempo[2];
  snap.B=tempo[3];
  snap.L=tempo[4];
  snap.H=tempo[5];
  if(type) {
    snap.PC=((word)tempo2[2])+256*((word)tempo2[3]);
    for(bucle=0;bucle<16;bucle++)
      snap.ay_regs[bucle]=tempo2[9+bucle];
    snap.ay_latch=tempo2[8];
  } else
    snap.PC=((word)tempo[6])+256*((word)tempo[7]);
  snap.SP=((word)tempo[8])+256*((word)tempo[9]);
  snap.I=tempo[10];
  snap.R=(tempo[11]&0x7F);

  if(tempo[12]==255)
    tempo[12]=1;

  if(tempo[12]&0x01)
    snap.R|=0x80;

  snap.border=(tempo[12]>>1)&0x07;

  if(tempo[12]&30)
    compressed=1;
  else
    compressed=0;

  snap.E=tempo[13];
  snap.D=tempo[14];
  snap.CC=tempo[15];
  snap.BB=tempo[16];
  snap.EE=tempo[17];
  snap.DD=tempo[18];
  snap.LL=tempo[19];
  snap.HH=tempo[20];
  snap.AA=tempo[21];
  snap.FF=tempo[22];
  snap.IY=((word)tempo[23])+256*((word)tempo[24]);
  snap.IX=((word)tempo[25])+256*((word)tempo[26]);
  if(tempo[27])
    snap.IFF1=1;
  else
    snap.IFF1=0;
  if(tempo[28])
    snap.IFF2=1;
  else
    snap.IFF2=0;
  switch(tempo[29]&0x03) {
  case 0:
    snap.Imode=0;
    break;
  case 1:
    snap.Imode=1;
    break;
  case 2:
    snap.Imode=2;
    break;
  }
  if(tempo[29]&0x04)
    snap.issue=2;
  else
    snap.issue=3;

  if(type)
    snap.pager=tempo2[5];

  if(type) { // extended z80
    if(snap.type==1) { // 128K snapshot

      /*       fclose(fichero);
	       return -3;*/  // z80 file not yet supported

      while(!feof(fichero)) {
	
	fread(byte_read,3,1,fichero);
	if(feof(fichero))
	  break;
	longitud2=((int)byte_read[0])+256*((int)byte_read[1]);
	switch(byte_read[2]) {
	case 3:
	  page=0;
	  break;
	case 4:
	  page=1;
	  break;
	case 5:
	  page=2;
	  break;
	case 6:
	  page=3;
	  break;
	case 7:
	  page=4;
	  break;
	case 8:
	  page=5;
	  break;
	case 9:
	  page=6;
	  break;
	case 10:
	  page=7;
	  break;
	default:
	  page=11;
	  break;
	}
	printf("Loading page %d of length %d\n",page,longitud);
	if(longitud2==0xFFFF) // uncompressed raw data
	  fread(snap.page[page],16384,1,fichero);
	else
	  uncompress_z80(fichero,16384,snap.page[page]);
      }

    } else {

      while(!feof(fichero)) {
	
	fread(byte_read,3,1,fichero);
	if(feof(fichero))
	  break;
	longitud2=((int)byte_read[0])+256*((int)byte_read[1]);
	switch(byte_read[2]) {
	case 8:
	  page=0;
	  break;
	case 4:
	  page=1;
	  break;
	case 5:
	  page=2;
	  break;
	default:
	  page=11;
	  break;
	}
	if(longitud2==0xFFFF) // uncompressed raw data
	  fread(snap.page[page],16384,1,fichero);
	else
	  uncompress_z80(fichero,16384,snap.page[page]);

      }
    }
  } else {

    if(compressed) {
      // 48k compressed z80 loader
      
      // we uncompress first the data
      
      uncompress_z80(fichero,49152,memo);
      
      memcpy(snap.page[0],memo,16384);
      memcpy(snap.page[1],memo+16384,16384);
      memcpy(snap.page[2],memo+32768,16384);
      
    } else {
      // 48k uncompressed z80 loader
      
      fread(snap.page[0],16384,1,fichero);
      fread(snap.page[1],16384,1,fichero);
      fread(snap.page[2],16384,1,fichero);
    }
  }

  load_snap(&snap);
  fclose(fichero);
  return 0; // all right
}


void load_snap(struct z80snapshot *snap) {

  int bucle;

  switch(snap->type) {
  case 0: // 48k
    ordenador.mode128k=0; // 48K mode
    ordenador.issue=snap->issue;
    ResetComputer();
    break;
  case 1: // 128k
    ordenador.mode128k=2; // +2 mode
    ordenador.issue=3;
    ResetComputer();
    printf("Pager: %X\n",snap->pager);
    OutZ80(0x7FFD,snap->pager);
    break;
  default:
    break;
  }

  procesador.AF.B.h=snap->A;
  procesador.AF.B.l=snap->F;
  procesador.BC.B.h=snap->B;
  procesador.BC.B.l=snap->C;
  procesador.DE.B.h=snap->D;
  procesador.DE.B.l=snap->E;
  procesador.HL.B.h=snap->H;
  procesador.HL.B.l=snap->L;
  procesador.AF1.B.h=snap->AA;
  procesador.AF1.B.l=snap->FF;
  procesador.BC1.B.h=snap->BB;
  procesador.BC1.B.l=snap->CC;
  procesador.DE1.B.h=snap->DD;
  procesador.DE1.B.l=snap->EE;
  procesador.HL1.B.h=snap->HH;
  procesador.HL1.B.l=snap->LL;
  procesador.IX.W=snap->IX;
  procesador.IY.W=snap->IY;
  procesador.SP.W=snap->SP;
  procesador.PC.W=snap->PC;
  procesador.I=snap->I;
  procesador.R=snap->R;

  if(snap->IFF1)
    procesador.IFF|=(IFF_1|IFF_EI);
  if(snap->IFF2)
    procesador.IFF|=IFF_2;

  switch(snap->Imode) {
  case 0:
    procesador.IFF&=~(IFF_IM1|IFF_IM2);
    break;
  case 1:
    procesador.IFF=(procesador.IFF&~IFF_IM2)|IFF_IM1;
    break;
  case 2:
    procesador.IFF=(procesador.IFF&~IFF_IM1)|IFF_IM2;
    break;
  }
  ordenador.border=snap->border+16;

  switch(snap->type) {
  case 0: // 48K

    for(bucle=0;bucle<16384;bucle++) {
      ordenador.memoria[bucle+147456]=snap->page[0][bucle];
      ordenador.memoria[bucle+98304]=snap->page[1][bucle];
      ordenador.memoria[bucle+114688]=snap->page[2][bucle];
    }
    
    ordenador.ay_emul=0;
    break;
  case 1: // 128K

    for(bucle=0;bucle<16384;bucle++) {
      ordenador.memoria[bucle+65536]=snap->page[0][bucle];
      ordenador.memoria[bucle+81920]=snap->page[1][bucle];
      ordenador.memoria[bucle+98304]=snap->page[2][bucle];
      ordenador.memoria[bucle+114688]=snap->page[3][bucle];
      ordenador.memoria[bucle+131072]=snap->page[4][bucle];
      ordenador.memoria[bucle+147456]=snap->page[5][bucle];
      ordenador.memoria[bucle+163840]=snap->page[6][bucle];
      ordenador.memoria[bucle+180224]=snap->page[7][bucle];
    }
    ordenador.ay_emul=1;
    for(bucle=0;bucle<16;bucle++)
      ordenador.ay_registers[bucle]=snap->ay_regs[bucle];
    ordenador.ay_latch=snap->ay_latch;
    break;
  default:
    break;
  }
}



