#include "Z80.h"
#include "computer.h"
#include "emulator.h"
#include "menus.h"
#include "characters.h"
#include <stdlib.h>
#include <string.h>


// loads the next block into memory

void fastload_block(FILE *fichero) {

  unsigned int longitud;
  unsigned char value[65536],salir;

  ordenador.other_ret=1; // next instruction must be RET

  if(!(procesador.AF.B.l&C_FLAG)) { // if Carry=0, is VERIFY, so return OK
    procesador.AF.B.l|=C_FLAG; // verify OK
    procesador.IX.W+=procesador.DE.W;
    procesador.DE.W=0;
    return;
  }

  if(fichero==NULL) {
    procesador.AF.B.l&=(~C_FLAG); // Load error
    procesador.IX.W+=procesador.DE.W;
    procesador.DE.W=0;
    sprintf(ordenador.osd_text,"No tape selected");
    ordenador.osd_time=50;
    return;
  }

  fread(value,2,1,fichero); // read length of current block
  if(feof(fichero)) { // end of file?
    sprintf(ordenador.osd_text,"Tape rewind");
    ordenador.osd_time=50;
    rewind(fichero); // again
    fread(value,2,1,fichero); // read length of current block
  }    
  longitud=((unsigned int)value[0])+256*((unsigned int)value[1]);
  fread(value,1,1,fichero); // read flag byte
  longitud--;
  if(value[0]!=procesador.AF.B.h) {
    fread(value,longitud,1,fichero); // jump to the next block
    return; // return without error
  }

  salir=0;
  do {
    if(longitud==0)
      salir=1;
    if(procesador.DE.W==0)
      salir=2;
    if(!salir) {
      fread(value,1,1,fichero); // read byte
      WrZ80(procesador.IX.W,(byte)value[0]); // store the byte
      procesador.IX.W++;
      if(procesador.IX.W>65535)
	procesador.IX.W=0;
      procesador.DE.W--;
      longitud--;
    }
  } while(!salir);

  clean_screen();

  if(salir==1) { // system wants to load more bytes that the existent
    procesador.AF.B.l&=(~C_FLAG); // Load error
    return;
  }

  if((salir==2)&&(longitud!=1)) { // there are more bytes to load
    procesador.AF.B.l&=(~C_FLAG); // Load error
    if(longitud>1)
      fread(value,longitud,1,fichero); // jump to the next block
    return;
  }

  fread(value,1,1,fichero); // jump over the checksum
  procesador.AF.B.l|=C_FLAG; // Load OK
  return;

}


/* reads a tape file and updates the readed bit */

inline void tape_read(FILE *fichero,int tstados) {

  static unsigned char value,value2;

  if(fichero==NULL)
    return;

  if(!ordenador.pause) {

    if(ordenador.tape_mode==3) { // initialize a new block
      fread(&value,1,1,fichero);
      fread(&value2,1,1,fichero); // read block longitude
      if(feof(fichero)) {
	rewind(fichero);
	ordenador.tape_mode=3;
	ordenador.pause=1;
	return;
      }
      ordenador.tape_byte_counter=((unsigned int)value)+256*((unsigned int)value2);
      fread(&(ordenador.tape_byte),1,1,fichero);
      ordenador.tape_bit=0x80;
      ordenador.tape_mode=0;
      ordenador.tape_counter0=2168;
      ordenador.tape_counter1=2168;    
      if(!(0x80&ordenador.tape_byte))
	ordenador.tape_counter_rep=3228; // 4 seconds
      else
	ordenador.tape_counter_rep=1614; // 2 seconds
      
    }
    
    // if there's a pulse still being reproduce, just reproduce it
    
    if(ordenador.tape_counter0) {
      if(ordenador.tape_counter0>tstados) {
	ordenador.tape_readed=0; // return 0
	ordenador.tape_counter0-=tstados; // decrement counter;
	return;
      } else {
	tstados-=ordenador.tape_counter0;
	ordenador.tape_counter0=0;
      }
    }
    
    ordenador.tape_readed=1; // return 1
    if(ordenador.tape_counter1) {
      if(ordenador.tape_counter1>tstados) {
	ordenador.tape_counter1-=tstados; // decrement counter;
	return;
      } else {
	tstados-=ordenador.tape_counter1;
	ordenador.tape_counter1=0;
	ordenador.tape_readed=0; // return 0
      }
    }
    
    // pulse ended

    switch(ordenador.tape_mode) {
    case 0: // guide tone
      if(ordenador.tape_counter_rep) { // still into guide tone
	ordenador.tape_counter_rep--;
	ordenador.tape_counter0=2168-tstados;
	ordenador.tape_counter1=2168; // new pulse
	return;
      } else { // guide tone ended. send sync tone
	ordenador.tape_counter0=667-tstados;
	ordenador.tape_counter1=735; // new pulse
	ordenador.tape_mode=1; // data mode
	ordenador.tape_bit=0x80; // from bit0 to bit7
	return;
      }
      break;
    case 1: // data
      if(ordenador.tape_byte&ordenador.tape_bit) { // next bit is 1
	ordenador.tape_counter0=1710-tstados;
	ordenador.tape_counter1=1710;
      } else {
	ordenador.tape_counter0=851-tstados;
	ordenador.tape_counter1=852;
      }
      ordenador.tape_bit=((ordenador.tape_bit>>1)&0x7F); // from bit0 to bit7
      if(!ordenador.tape_bit) {
	ordenador.tape_byte_counter--;
	if(!ordenador.tape_byte_counter) { // ended the block
	  ordenador.tape_mode=2; // pause
	  ordenador.tape_counter_rep=3500000; // 1 seconds
	  return;
	}
	ordenador.tape_bit=0x80;
	fread(&(ordenador.tape_byte),1,1,fichero); // read next byte
	if(feof(fichero)) {
	  rewind(fichero);
	  ordenador.tape_mode=4; // stop tape
	  return;
	}
      }
      break;
    case 2: // pause
      ordenador.tape_readed=0;
      if(ordenador.tape_counter_rep>tstados) {
	ordenador.tape_counter_rep-=tstados;
      } else {
	ordenador.tape_counter_rep=0;
	ordenador.tape_mode=3; // read new block
      }
      return;
      break;
    case 4:
      ordenador.tape_mode=3; // initialize
      ordenador.pause=1; // pause it
      break;
    }
  }

}

/* Returns the bus value when reading a port without a periferial */

inline byte bus_empty() {

  /*  ordenador.bus_counter++;

  if(ordenador.bus_counter==10) {
    ordenador.bus_counter=0;
    return 191;
  } else
  return 255; */

  if(ordenador.mode128k!=3)
    return (ordenador.bus_value);
  else
    return(255); // +2A and +3 returns always 255
}

/* calls all the routines that emulates the computer, runing them for 'tstados'
   tstates */

inline void emulate(int tstados) {

  show_screen(tstados);
  play_ay(tstados);
  play_sound(tstados);
  tape_read(ordenador.tap_file,tstados);
  if(!ordenador.pause) {
    if(ordenador.tape_readed)
      ordenador.sound_bit=1;
    else
      ordenador.sound_bit=0;// if not paused, asign SOUND_BIT the value of tape
  }
}

void computer_init() {

  int bucle;

  ordenador.bus_counter=0;
  ordenador.port254=0;
  ordenador.issue=2;
  ordenador.mode128k=0;
  ordenador.joystick=0;

  ordenador.tape_readed=0;
  ordenador.pause=1; // tape stop
  ordenador.fast_load=1; // fast load by default
  ordenador.tape_mode=3;
  ordenador.tap_file=NULL;

  ordenador.osd_text[0]=0;
  ordenador.osd_time=0;

  ordenador.other_ret=0;

  ordenador.s8=ordenador.s9=ordenador.s10=ordenador.s11=ordenador.s12=ordenador.s13=ordenador.s14=ordenador.s15=0xFF;

  ordenador.js=0x00;

  for(bucle=0;bucle<16;bucle++)
    ordenador.ay_registers[bucle]=0;
  ordenador.ay_emul=0;
  ordenador.aych_a=0;
  ordenador.aych_b=0;
  ordenador.aych_c=0;
  ordenador.aych_n=0;
  ordenador.aych_envel=0;
  ordenador.vol_a=0;
  ordenador.vol_b=0;
  ordenador.vol_c=0;
  ordenador.tst_ay=0;
  ordenador.tst_ay2=0;

  ordenador.ayval_a=0;
  ordenador.ayval_b=0;
  ordenador.ayval_c=0;
  ordenador.ayval_n=0;
  ordenador.ay_envel_value=0;
  ordenador.ay_envel_way=0;
}


/* Registers the screen surface where the Spectrum will put the picture,
prepares the palette and creates two arrays (translate and translate2)
that gives the memory address for each scan */

void register_screen(SDL_Surface *pantalla) {

  int resx,resy,bucle,bucle2,bucle3,bucle4,bucle5;
  SDL_Color colores[16];

  // we prepare the scanline transform arrays

  bucle5=0;
  for(bucle=0;bucle<3;bucle++)
    for(bucle2=0;bucle2<8;bucle2++)
      for(bucle3=0;bucle3<8;bucle3++)
	for(bucle4=0;bucle4<32;bucle4++) {
	  ordenador.translate[bucle5]=147456+bucle*2048+bucle2*32+bucle3*256+bucle4;
	  ordenador.translate2[bucle5]=153600+bucle*256+bucle2*32+bucle4;
	  bucle5++;
	}
  ordenador.tstados_counter=0;

  ordenador.screen=pantalla;

  colores[0].r=0;
  colores[0].g=0;
  colores[0].b=0;
  colores[1].r=0;
  colores[1].g=0;
  colores[1].b=192;
  colores[2].r=192;
  colores[2].g=0;
  colores[2].b=0;
  colores[3].r=192;
  colores[3].g=0;
  colores[3].b=192;
  colores[4].r=0;
  colores[4].g=192;
  colores[4].b=0;
  colores[5].r=0;
  colores[5].g=192;
  colores[5].b=192;
  colores[6].r=192;
  colores[6].g=192;
  colores[6].b=0;
  colores[7].r=192;
  colores[7].g=192;
  colores[7].b=192;
  colores[8].r=0;
  colores[8].g=0;
  colores[8].b=0;
  colores[9].r=0;
  colores[9].g=0;
  colores[9].b=255;
  colores[10].r=255;
  colores[10].g=0;
  colores[10].b=0;
  colores[11].r=255;
  colores[11].g=0;
  colores[11].b=255;
  colores[12].r=0;
  colores[12].g=255;
  colores[12].b=0;
  colores[13].r=0;
  colores[13].g=255;
  colores[13].b=255;
  colores[14].r=255;
  colores[14].g=255;
  colores[14].b=0;
  colores[15].r=255;
  colores[15].g=255;
  colores[15].b=255;

  SDL_SetColors(ordenador.screen,colores,16,16); // set 16 colors from the 16th

  ordenador.border=0;
  ordenador.currline=0;
  ordenador.currpix=0;
  ordenador.flash=0;

  resx=ordenador.screen->w;
  resy=ordenador.screen->h;
  // bpp=screen->format->BytesPerPixel;

  // we admit only 640x480 and 800x600

  switch(resx) {
  case 640:
    ordenador.init_line=0;
    ordenador.next_line=640;
    ordenador.first_line=40;
    ordenador.last_line=280;
    ordenador.first_pixel=16;
    ordenador.last_pixel=336;
    break;
  case 800:
    ordenador.init_line=48;
    ordenador.next_line=896;
    ordenador.first_line=6;
    ordenador.last_line=305;
    ordenador.first_pixel=0;
    ordenador.last_pixel=351;
    break;
  }
  // ordenador.init_line*=bpp;
  // ordenador.next_line*=bpp;
  ordenador.pixel=((unsigned char *)(ordenador.screen->pixels))+ordenador.init_line;
  ordenador.interr=0;

  ordenador.p_translt=ordenador.translate;
  ordenador.p_translt2=ordenador.translate2;

  ordenador.contador_flash=0;
  ordenador.readed=0;

  ordenador.tstados_counter_sound=0;
  ordenador.current_buffer=sound1;
  ordenador.num_buff=0; // first buffer
  ordenador.sound_cuantity=0;
  ordenador.pixancho=447;
  ordenador.pixalto=311; // values for 48K mode
}


void fill_audio(void *udata,Uint8 *stream,int len) {

  static int bucle;
  // static struct sound_buffer *snd_temp;
  static Uint8 *tempo,*tempo2;

  // snd_temp=(struct sound_buffer *)udata;
  tempo=stream;

  tempo2=(Uint8 *)snd_buffer.buffer;
  for(bucle=0;bucle<len;bucle++)
    *(tempo++)=*(tempo2++);
  SDL_mutexV(semaforo); // unlocks the main thread
  // printf("llamado %d\n",len);
}



/* emulates the AY-3-8912 during TSTADOS tstates */

inline void play_ay(unsigned int tstados) {

  if(!ordenador.ay_emul)
    return;

  ordenador.tst_ay+=tstados;
  ordenador.tst_ay2+=tstados;

  if(ordenador.tst_ay2>255) {
    ordenador.tst_ay2-=256;
    if((ordenador.ay_registers[11])||(ordenador.ay_registers[12])) {
      if(ordenador.aych_envel)
	ordenador.aych_envel--;
      else {
	ordenador.aych_envel=(((unsigned int)ordenador.ay_registers[11])+256*((unsigned int)(ordenador.ay_registers[12])));
	if(ordenador.ay_envel_way&0x02) // start cycle?
	  switch((ordenador.ay_registers[13])&0x0F) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  case 8:
	  case 9:
	  case 10:
	  case 11:
	    ordenador.ay_envel_way=4; // cycle started and decrementing
	    ordenador.ay_envel_value=16;
	    break;
	  default:
	    ordenador.ay_envel_way=5; // cycle started and incrementing
	    ordenador.ay_envel_value=-1;
	  }
	if(ordenador.ay_envel_way&0x04) { // cycle started?
	  switch((ordenador.ay_registers[13])&0x0F) {
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  case 9:
	    ordenador.ay_envel_value--;
	    if(ordenador.ay_envel_value==0)
	      ordenador.ay_envel_way=0; // end
	    break;

	  case 4:
	  case 5:
	  case 6:
	  case 7:
	  case 15:
	    ordenador.ay_envel_value++;
	    if(ordenador.ay_envel_value==16) {
	      ordenador.ay_envel_value=0;
	      ordenador.ay_envel_way=0; // end
	    }
	    break;

	  case 8:
	    ordenador.ay_envel_value--;
	    if(ordenador.ay_envel_value==-1)
	      ordenador.ay_envel_value=15; // repeat
	    break;

	  case 10:
	  case 14:
	    if(ordenador.ay_envel_way&0x01)
	      ordenador.ay_envel_value++;
	    else
	      ordenador.ay_envel_value--;
	    if(ordenador.ay_envel_value==16) {
	      ordenador.ay_envel_value=14;
	      ordenador.ay_envel_way=4;
	    }
	    if(ordenador.ay_envel_value==-1) {
	      ordenador.ay_envel_value=1;
	      ordenador.ay_envel_way=5;
	    }
	    break;

	  case 11:
	    ordenador.ay_envel_value--;
	    if(ordenador.ay_envel_value==-1) {
	      ordenador.ay_envel_value=15;
	      ordenador.ay_envel_way=0; // end
	    }
	    break;

	  case 12:
	    ordenador.ay_envel_value++;
	    if(ordenador.ay_envel_value==16)
	      ordenador.ay_envel_value=0;
	    break;

	  case 13:
	    ordenador.ay_envel_value++;
	    if(ordenador.ay_envel_value==15)
	      ordenador.ay_envel_way=0; // end
	    break;
	  }
	}
      }
    } else
      ordenador.ay_envel_value=15;
  }

  while(ordenador.tst_ay>=16) {
    ordenador.tst_ay-=16;

    if((ordenador.ay_registers[0])||(ordenador.ay_registers[1])) {
      if(ordenador.aych_a)
	ordenador.aych_a--;
      else {
	ordenador.ayval_a=1-ordenador.ayval_a;
	ordenador.aych_a=(((unsigned int)ordenador.ay_registers[0])+256*((unsigned int)((ordenador.ay_registers[1])&0x0F)))/2;
      }
    } else
      ordenador.ayval_a=0;

    if((ordenador.ay_registers[2])||(ordenador.ay_registers[3])) {
      if(ordenador.aych_b)
	ordenador.aych_b--;
      else {
	ordenador.ayval_b=1-ordenador.ayval_b;
	ordenador.aych_b=(((unsigned int)ordenador.ay_registers[2])+256*((unsigned int)((ordenador.ay_registers[3])&0x0F)))/2;
      }
    }else
	ordenador.ayval_b=0;

    if((ordenador.ay_registers[4])||(ordenador.ay_registers[5])) {
      if(ordenador.aych_c)
	ordenador.aych_c--;
      else {
	ordenador.ayval_c=1-ordenador.ayval_c;
	ordenador.aych_c=(((unsigned int)ordenador.ay_registers[4])+256*((unsigned int)((ordenador.ay_registers[5])&0x0F)))/2;
      }
    } else
      ordenador.ayval_c=0;

    if(ordenador.ay_registers[6]) {
      if(ordenador.aych_n)
	ordenador.aych_n--;
      else {
	ordenador.ayval_n=1-ordenador.ayval_n;
	ordenador.aych_n=((((unsigned int)ordenador.ay_registers[6])&0x1F)+(rand()%3))/2;
      if(ordenador.aych_n>16)
	ordenador.aych_n=16;
      }
    } else
      ordenador.ayval_n=0;

    if(ordenador.ay_registers[8]&0x10)
      ordenador.vol_a=(unsigned char)((((unsigned int)ordenador.ay_envel_value))*(unsigned int)ordenador.volume)/15;
    else
      ordenador.vol_a=(unsigned char)((((unsigned int)(ordenador.ay_registers[8]&0x0F))*(unsigned int)ordenador.volume)/15);

    if(ordenador.ay_registers[10]&0x10)
      ordenador.vol_c=(unsigned char)((((unsigned int)ordenador.ay_envel_value))*(unsigned int)ordenador.volume)/15;
    else
      ordenador.vol_c=(unsigned char)((((unsigned int)(ordenador.ay_registers[10]&0x0F))*(unsigned int)ordenador.volume)/15);

    if(ordenador.ay_registers[9]&0x10)
      ordenador.vol_b=(unsigned char)((((unsigned int)ordenador.ay_envel_value))*(unsigned int)ordenador.volume)/15;
    else
      ordenador.vol_b=(unsigned char)((((unsigned int)(ordenador.ay_registers[9]&0x0F))*(unsigned int)ordenador.volume)/15);

  }
}


/* Creates the sound buffer during the TSTADOS tstate that the Z80 used to
   execute last instruction */

inline void play_sound(unsigned int tstados) {

  static int bucle;
  static int value;
  static char sample_v;

  ordenador.tstados_counter_sound+=tstados;

  while(ordenador.tstados_counter_sound>=ordenador.tst_sample) {

    ordenador.tstados_counter_sound-=ordenador.tst_sample;

    for(bucle=0;bucle<ordenador.increment;bucle++) {
      sample_v=ordenador.sample1b[bucle];
      if((ordenador.sound_bit)&&(sample_v))
	value=ordenador.volume;
      else
	value=0;
      if(ordenador.ay_emul) { // if emulation is ON, emulate it
	if((ordenador.ayval_a)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x01)))
	  value+=(int)ordenador.vol_a;
	if((ordenador.ayval_b)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x02)))
	  value+=(int)ordenador.vol_b;	
	if((ordenador.ayval_c)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x04)))
	  value+=(int)ordenador.vol_c;
	   if((ordenador.ayval_n)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x08)))
	  value+=(int)ordenador.vol_a;
	if((ordenador.ayval_n)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x10)))
	  value+=(int)ordenador.vol_b;	
	if((ordenador.ayval_n)&&(sample_v)&&(!(ordenador.ay_registers[7]&0x20)))
	  value+=(int)ordenador.vol_c;
      }
      if(value>255)
	value=255;
      *ordenador.current_buffer=((char)value)+ordenador.sign;
      ordenador.current_buffer++;
    }
    ordenador.sound_cuantity++;
    if(ordenador.sound_cuantity==ordenador.buffer_len) { // buffer filled
      SDL_LockAudio();
      if(ordenador.num_buff) {
	ordenador.current_buffer=sound1;
	snd_buffer.buffer=sound2;
	ordenador.num_buff=0;
      } else {
	ordenador.current_buffer=sound2;
	snd_buffer.buffer=sound1;
	ordenador.num_buff=1;
      }
      ordenador.sound_cuantity=0;
      SDL_UnlockAudio();
      SDL_mutexP(semaforo); // lock it, so we wait to the thread to liberate it
    }
  }
}

/* Paints the spectrum screen during the TSTADOS tstates that the Z80 used
to execute last instruction */

inline void show_screen(int tstados) {

  static unsigned char temporal,ink,paper,fflash;

  ordenador.tstados_counter+=tstados;

  while(ordenador.tstados_counter>3) {
    ordenador.tstados_counter-=4;


    // test if current pixel is for border or for user area
    
    if((ordenador.currline<64)||(ordenador.currline>255)||(ordenador.currpix<48)||(ordenador.currpix>303)) {
      
      // is border
      paint_pixels(255,ordenador.border,0); // paint 8 pixels with BORDER color

      ordenador.bus_value=255;
      
    } else {
      
      // is user area. We search for ink and paper colours
      
      temporal=ordenador.memoria[(*ordenador.p_translt2)+ordenador.video_offset]; // attributes
      ordenador.bus_value=temporal;
      ink=temporal&0x07; // ink colour
      paper=(temporal>>3)&0x07; // paper colour
      if(temporal&0x40) { // bright flag?
	ink+=24;
	paper+=24;
      } else {
	ink+=16;
	paper+=16;
      }
      fflash=temporal&0x80; // flash flag
      temporal=ordenador.memoria[(*ordenador.p_translt)+ordenador.video_offset]; // data
      ordenador.p_translt++;
      ordenador.p_translt2++;
      if((fflash)&&(ordenador.flash))
	paint_pixels(temporal,paper,ink); // if FLASH, invert PAPER and INK
      else
	paint_pixels(temporal,ink,paper);
    }
    ordenador.currpix+=8;
    if(ordenador.currpix>ordenador.pixancho) {
      ordenador.currpix=0;
      ordenador.currline++;
      if(ordenador.currline>40) // ordenador.first_line)
	ordenador.pixel+=640; // ordenador.next_line;
      if(ordenador.currline>ordenador.pixalto) {
	if(ordenador.osd_time) {
	  ordenador.osd_time--;
	  if(ordenador.osd_time)
	    printstring(ordenador.screenbuffer,ordenador.osd_text,-1,232,28,16,ordenador.screen_width);
	  else
	    printstring(ordenador.screenbuffer,"                                        ",-1,232,28,16,ordenador.screen_width);
	    
	}

	if(ordenador.mustlock) {
	  SDL_UnlockSurface(ordenador.screen);
	  SDL_Flip(ordenador.screen);
	  SDL_LockSurface(ordenador.screen);
	} else
	  SDL_Flip(ordenador.screen);

	ordenador.currline=0;
	ordenador.interr=1;
	read_keyboard(); // read the physical keyboard
	ordenador.pixel=((unsigned char *)(ordenador.screen->pixels)); // +ordenador.init_line;
	ordenador.p_translt=ordenador.translate;
	ordenador.p_translt2=ordenador.translate2;
	ordenador.contador_flash++;
	if(ordenador.contador_flash==16) {
	  ordenador.flash=1-ordenador.flash;
	  ordenador.contador_flash=0;
	}
	// ordenador.readed=1; // we can now read again the keyboard if needed
      }
    } 
  }
}


/* PAINT_PIXELS paints one byte with INK colour for 1 bits and PAPER colour
for 0 bits, and increment acordingly the pointer PIXEL */

inline void paint_pixels(unsigned char octet,char ink,char paper) {

  static int bucle;
  static unsigned char mask,valor;


  /* if((ordenador.currpix<ordenador.first_pixel)||(ordenador.currpix>=ordenador.last_pixel)||(ordenador.currline<ordenador.first_line)||(ordenador.currline>=ordenador.last_line))
     return; */
  if((ordenador.currpix<16)||(ordenador.currpix>=336)||(ordenador.currline<40)||(ordenador.currline>=280))
    return;

  if((ordenador.osd_time)&&(ordenador.currline>152)&&(ordenador.currline<168)){
    ordenador.pixel+=16;
    return;
  }

  mask=0x80;
  for(bucle=0;bucle<8;bucle++) {
    valor=(octet&mask) ? ink : paper;
    *ordenador.pixel=valor;
    ordenador.pixel++;
    *ordenador.pixel=valor;
    ordenador.pixel++;
    mask=((mask>>1)&0x7F);
  }
}
    

// Read the keyboard and stores the flags

inline void read_keyboard() {

  // Uint8 *keystate;

  static unsigned int temporal_io;
  static SDL_Event evento;

  ordenador.k8=ordenador.k9=ordenador.k10=ordenador.k11=ordenador.k12=ordenador.k13=ordenador.k14=ordenador.k15=ordenador.jk=0;

  if(!SDL_PollEvent(&evento))
    return;

  if((evento.type!=SDL_KEYDOWN)&&(evento.type!=SDL_KEYUP))
    return;

  temporal_io=(unsigned int)evento.key.keysym.sym;


  if(evento.type==SDL_KEYUP)
    switch(temporal_io) {
    case SDLK_ESCAPE: // to exit from the emulator
      salir=0;
      return;
      break;
      
    case SDLK_F1:
      SDL_PauseAudio(1);
      help_menu(); // shows the help menu
      SDL_PauseAudio(0);
      break;

    case SDLK_F2:
      SDL_PauseAudio(1);
      snapshots_menu(); // manage snapshot files
      SDL_PauseAudio(0);
      break;

    case SDLK_F3:
      SDL_PauseAudio(1);
      taps_menu(); // manage TAP files
      SDL_PauseAudio(0);
      break;

    case SDLK_F4: // settings
      SDL_PauseAudio(1);
      settings_menu();
      SDL_PauseAudio(0);
      break;
    case SDLK_F5: // STOP tape
      if(ordenador.fast_load==0)
	ordenador.pause=1;
      break;
    case SDLK_F6: // PLAY tape
      if(ordenador.fast_load==0)      
	ordenador.pause=0;
      break;


    case SDLK_F10: // Reset emulator
      ResetComputer();
      ordenador.pause=1;
      if(ordenador.tap_file!=NULL) {
	ordenador.tape_mode=3;
	rewind(ordenador.tap_file);
      }
      break;
      
    case SDLK_F11: // lower volume
      if(ordenador.volume>3)
	set_volume(ordenador.volume-4);
      sprintf(ordenador.osd_text," Volume: %d ",ordenador.volume/4);
      ordenador.osd_time=50;
      break;
    case SDLK_F12: // upper volume
      set_volume(ordenador.volume+4);
      sprintf(ordenador.osd_text," Volume: %d ",ordenador.volume/4);
      ordenador.osd_time=50;
      break;
    }


  // test for joystick

  switch(temporal_io) {
  case SDLK_UP:
    switch(ordenador.joystick) {
    case 0: // cursor
      temporal_io=SDLK_7;
      break;
    case 1:
      ordenador.jk=8;
      break;
    case 2: // sinclair 1
      temporal_io=SDLK_4;
      break;
    break;
    case 3: // sinclair 2
      temporal_io=SDLK_7;
      break;
    break;
    }
    break;
  case SDLK_DOWN:
    switch(ordenador.joystick) {
    case 0: // cursor
      temporal_io=SDLK_6;
      break;
    case 1:
      ordenador.jk=4;
      break;
    case 2: // sinclair 1
      temporal_io=SDLK_3;
      break;
    break;
    case 3: // sinclair 2
      temporal_io=SDLK_8;
      break;
    break;
    }
    break;
  case SDLK_RIGHT:
    switch(ordenador.joystick) {
    case 0: // cursor
      temporal_io=SDLK_8;
      break;
    case 1:
      ordenador.jk=1;
      break;
    case 2: // sinclair 1
      temporal_io=SDLK_1;
      break;
    break;
    case 3: // sinclair 2
      temporal_io=SDLK_9;
      break;
    break;
    }
    break;
  case SDLK_LEFT:
    switch(ordenador.joystick) {
    case 0: // cursor
      temporal_io=SDLK_5;
      break;
    case 1:
      ordenador.jk=2;
      break;
    case 2: // sinclair 1
      temporal_io=SDLK_2;
      break;
    break;
    case 3: // sinclair 2
      temporal_io=SDLK_0;
      break;
    break;
    }
    break;
  case SDLK_RALT:
  case SDLK_LALT:
    switch(ordenador.joystick) {
    case 0: // cursor
      temporal_io=SDLK_0;
      break;
    case 1:
      ordenador.jk=16;
      break;
    case 2: // sinclair 1
      temporal_io=SDLK_5;
      break;
    break;
    case 3: // sinclair 2
      temporal_io=SDLK_6;
      break;
    break;
    }
    break;
  }

  switch(temporal_io) {

  case SDLK_SPACE:
    ordenador.k15=1;
    break;
  case SDLK_RCTRL:
  case SDLK_LCTRL:
    ordenador.k15=2;
    break;
  case SDLK_m:
    ordenador.k15=4;
    break;
  case SDLK_n:
    ordenador.k15=8;
    break;
  case SDLK_b:
    ordenador.k15=16;
    break;
  case SDLK_RETURN:
    ordenador.k14=1;
    break;
  case SDLK_l:
    ordenador.k14=2;
    break;
  case SDLK_k:
    ordenador.k14=4;
    break;
  case SDLK_j:
    ordenador.k14=8;
    break;
  case SDLK_h:
    ordenador.k14=16;
    break;
  case SDLK_p:
    ordenador.k13=1;
    break;
  case SDLK_o:
    ordenador.k13=2;
    break;
  case SDLK_i:
    ordenador.k13=4;
    break;
  case SDLK_u:
    ordenador.k13=8;
    break;
  case SDLK_y:
    ordenador.k13=16;
    break;
  case SDLK_0:
    ordenador.k12=1;
    break;
  case SDLK_9:
    ordenador.k12=2;
    break;
  case SDLK_8:
    ordenador.k12=4;
    break;
  case SDLK_7:
    ordenador.k12=8;
    break;
  case SDLK_6:
    ordenador.k12=16;
    break;
  case SDLK_1:
    ordenador.k11=1;
    break;
  case SDLK_2:
    ordenador.k11=2;
    break;
  case SDLK_3:
    ordenador.k11=4;
    break;
  case SDLK_4:
    ordenador.k11=8;
    break;
  case SDLK_5:
    ordenador.k11=16;
    break;
  case SDLK_q:
    ordenador.k10=1;
    break;
  case SDLK_w:
    ordenador.k10=2;
    break;
  case SDLK_e:
    ordenador.k10=4;
    break;
  case SDLK_r:
    ordenador.k10=8;
    break;
  case SDLK_t:
    ordenador.k10=16;
    break;
  case SDLK_a:
    ordenador.k9=1;
    break;
  case SDLK_s:
    ordenador.k9=2;
    break;
  case SDLK_d:
    ordenador.k9=4;
    break;
  case SDLK_f:
    ordenador.k9=8;
    break;
  case SDLK_g:
    ordenador.k9=16;
    break;
  case SDLK_RSHIFT:
  case SDLK_LSHIFT:
    ordenador.k8=1;
    break;
  case SDLK_z:
    ordenador.k8=2;
    break;
  case SDLK_x:
    ordenador.k8=4;
    break;
  case SDLK_c:
    ordenador.k8=8;
    break;
  case SDLK_v:
    ordenador.k8=16;
    break;
  }

  if(evento.type==SDL_KEYUP) {
    ordenador.s8|=ordenador.k8;
    ordenador.s9|=ordenador.k9;
    ordenador.s10|=ordenador.k10;
    ordenador.s11|=ordenador.k11;
    ordenador.s12|=ordenador.k12;
    ordenador.s13|=ordenador.k13;
    ordenador.s14|=ordenador.k14;
    ordenador.s15|=ordenador.k15;
    ordenador.js&=(ordenador.jk^255);
  } else {
    ordenador.s8&=(ordenador.k8^255);
    ordenador.s9&=(ordenador.k9^255);
    ordenador.s10&=(ordenador.k10^255);
    ordenador.s11&=(ordenador.k11^255);
    ordenador.s12&=(ordenador.k12^255);
    ordenador.s13&=(ordenador.k13^255);
    ordenador.s14&=(ordenador.k14^255);
    ordenador.s15&=(ordenador.k15^255);
    ordenador.js|=ordenador.jk;
  }

  return;
}

// resets the computer and loads the right ROMs

void ResetComputer() {

  static int bucle;

  ResetZ80(&procesador);
  load_rom(ordenador.mode128k);

  // reset the AY-3-8912

  for(bucle=0;bucle<16;bucle++)
    ordenador.ay_registers[bucle]=0;
  
  ordenador.aych_a=0;
  ordenador.aych_b=0;
  ordenador.aych_c=0;
  ordenador.aych_n=0;
  ordenador.aych_envel=0;
  ordenador.vol_a=0;
  ordenador.vol_b=0;
  ordenador.vol_c=0;

  ordenador.mport1=0;
  ordenador.mport2=0;
  ordenador.video_offset=0; // video in page 9 (page 5 in 128K)
  switch(ordenador.mode128k) {
  case 0: // 48K

    ordenador.pixancho=447;
    ordenador.pixalto=311;

    ordenador.block0=ordenador.memoria;
    ordenador.block1=ordenador.memoria+131072; // video mem. in page 9 (page 5 in 128K)
    ordenador.block2=ordenador.memoria+65536; // 2nd block in page 6 (page 2 in 128K)
    ordenador.block3=ordenador.memoria+65536; // 3rd block in page 7 (page 3 in 128K)
    ordenador.mport1=32; // access to port 7FFD disabled
    break;
  case 3: // +2A/+3
    OutZ80(0x1FFD,0);
  case 1: // 128K
  case 2: // +2
    OutZ80(0x7FFD,0);
    ordenador.pixancho=455;
    ordenador.pixalto=310;
    break;
  }
  
}

void WrZ80(register word Addr,register byte Value) {
  
  switch(Addr&0xC000) {
  case 0x0000:
    // only writes below the first 16K if we are in +3 mode and bit0 of mport2 is 1
    
    if((ordenador.mode128k==3)&&(1==(ordenador.mport2&0x01)))
      *(ordenador.block0+Addr)=(unsigned char) Value;
    break;
  case 0x4000:
    *(ordenador.block1+Addr)=(unsigned char) Value;
    break;
  case 0x8000:
    *(ordenador.block2+Addr)=(unsigned char) Value;
    break;
  case 0xC000:
    *(ordenador.block3+Addr)=(unsigned char) Value;
    break;
  }
}


byte RdZ80(register word Addr) {
  
  switch(ordenador.other_ret) {
  case 1:
    ordenador.other_ret=0;
    return (201); // RET instruction
    break;
  default:
    switch(Addr&0xC000) {
    case 0x0000:
      return ((byte)(*(ordenador.block0+Addr)));
      break;
    case 0x4000:
      return ((byte)(*(ordenador.block1+Addr)));
      break;
    case 0x8000:
      return ((byte)(*(ordenador.block2+Addr)));
      break;
    case 0xC000:
      return ((byte)(*(ordenador.block3+Addr)));
      break;
    default:
      printf("Memory error\n");
      exit(1);
      return 0;
    }
    break;
  }
}

void set_memory_pointers() {

  static unsigned int rom,ram;

  // assign the offset for video page

  if(ordenador.mport1&0x08)
    ordenador.video_offset=32768; // page 7
  else
    ordenador.video_offset=0; // page 5

  // assign ROMs and, if in special mode, RAM for the whole blocks

  if((ordenador.mode128k==3)) {
    if(ordenador.mport2&0x01) { // +2A/+3 special mode
      ram=(unsigned int)(ordenador.mport1&0x06); // bits 1&2
      switch(ram) {
      case 0:
	ordenador.block0=ordenador.memoria+65536;
	ordenador.block1=ordenador.memoria+65536;
	ordenador.block2=ordenador.memoria+65536;
	ordenador.block3=ordenador.memoria+65536;
	break;
      case 2:
	ordenador.block0=ordenador.memoria+131072;
	ordenador.block1=ordenador.memoria+131072;
	ordenador.block2=ordenador.memoria+131072;
	ordenador.block3=ordenador.memoria+131072;
	break;
      case 4:
	ordenador.block0=ordenador.memoria+131072;
	ordenador.block1=ordenador.memoria+131072;
	ordenador.block2=ordenador.memoria+131072;
	ordenador.block3=ordenador.memoria+65536;
	break;
      case 6:
	ordenador.block0=ordenador.memoria+131072;
	ordenador.block1=ordenador.memoria+163840;
	ordenador.block2=ordenador.memoria+131072;
	ordenador.block3=ordenador.memoria+65536;
	break;
      }
      return;
    } else { // ROMs for +2A/+3 normal mode
      rom=0;
      if(ordenador.mport1&0x10)
	rom++;
      if(ordenador.mport2&0x04)
	rom+=2;
      // assign the first block pointer to the right block bank
      ordenador.block0=ordenador.memoria+(16384*rom);
    }
  } else { // ROMs for 128K/+2 mode
    if(ordenador.mport1&0x10)
      ordenador.block0=ordenador.memoria+16384;
    else
      ordenador.block0=ordenador.memoria;
  }

  // RAMs for 128K/+2 mode, and +2A/+3 in normal mode

  ordenador.block1=ordenador.memoria+131072; // page 5 minus 16384
  ordenador.block2=ordenador.memoria+65536; // page 2 minus 32768
  
  ram=1+((unsigned int)(ordenador.mport1&0x07)); // RAM page for block3 plus 1
  ordenador.block3=ordenador.memoria+(16384*ram); // page n minus 49152

  
  // printf("pager: %x\noffset: %d\nblock 0: %d\nblock 1: %d\nblock2: %d\nblock3: %d\n",ordenador.mport1,ordenador.video_offset,ordenador.block0-ordenador.memoria,ordenador.block1-ordenador.memoria,ordenador.block2-ordenador.memoria,ordenador.block3-ordenador.memoria);

}


void OutZ80(register word Port, register byte Value) {
  
  // ULA port (A0 low)
  
  if(!(Port&0x0001)) {

    ordenador.port254=(unsigned char)Value;

    ordenador.border=16+(((unsigned char)Value)&0x07);

    if(ordenador.pause) {
      if(Value&0x10)
	ordenador.sound_bit=1;
      else
	ordenador.sound_bit=0; // assign to SOUND_BIT the value
    }
  }

  // Memory page (7FFD & 1FFD)

  if((Port==0x7FFD)&&(0==(ordenador.mport1&0x20))) {
    ordenador.mport1=(unsigned char) Value;
    set_memory_pointers(); // set the pointers
  }

  if((Port==0x1FFD)&&(0==(ordenador.mport1&0x20))) {
    ordenador.mport2=(unsigned char) Value;
    set_memory_pointers(); // set the pointers
  }

  // Sound chip (AY-3-8912)
  
  if(Port==0xFFFD)
    ordenador.ay_latch=((unsigned int) (Value&0x0F));
  
  if(Port==0xBFFD) {
    ordenador.ay_registers[ordenador.ay_latch]=(unsigned char) Value;
    if(ordenador.ay_latch==13)
      ordenador.ay_envel_way=2; // start cycle
  }
  
}


byte InZ80(register word Port) {

  static unsigned int temporal_io;
  byte pines;

  temporal_io=(unsigned int) Port;

  if(!(temporal_io&0x0001)) {
    
    pines=0xBF; // by default, sound bit is 0
    if(!(temporal_io&0x0100))
      pines&=ordenador.s8;
    if(!(temporal_io&0x0200))
      pines&=ordenador.s9;
    if(!(temporal_io&0x0400))
      pines&=ordenador.s10;
    if(!(temporal_io&0x0800))
      pines&=ordenador.s11;
    if(!(temporal_io&0x1000))
      pines&=ordenador.s12;
    if(!(temporal_io&0x2000))
      pines&=ordenador.s13;
    if(!(temporal_io&0x4000))
      pines&=ordenador.s14;
    if(!(temporal_io&0x8000))
      pines&=ordenador.s15;

    if(ordenador.pause) {
      if(ordenador.issue==2) {
	if(ordenador.port254&0x18)
	  pines|=0x40;
      } else {
	if(ordenador.port254&0x10)
	  pines|=0x40;
      }
    } else {
      if(ordenador.tape_readed)
	pines|=0x40; // sound input
      else
	pines&=0xBF; // sound input
    }

    return(pines);
  }

  if((!(temporal_io&0x0020))&&(ordenador.joystick==1))
    return(ordenador.js);

  if(temporal_io==0xFFFD)
    return(ordenador.ay_registers[ordenador.ay_latch]);

  return(bus_empty());

}

void PatchZ80(register Z80 *R) {

}

word LoopZ80(register Z80 *R) {

  if(ordenador.interr==0)
    return (INT_NONE);
  else
    return (0x0038);
}


void set_volume(unsigned char volume) {

  unsigned char vol2;
  int bucle;

  if(volume>40)
    vol2=40;
  else
    vol2=volume;

  ordenador.volume=vol2;

  for(bucle=0;bucle<4;bucle++) {
    ordenador.sample0[bucle]=0;
    ordenador.sample1[bucle]=0;
    ordenador.sample1b[bucle]=0;
  }

  switch(ordenador.format) {
  case 0:
    ordenador.sample1[0]=2*vol2;
    ordenador.sample1[1]=2*vol2;
    ordenador.sample1b[0]=1;
    ordenador.sample1b[1]=1;
    break;
  case 1:
    ordenador.sample1[1]=2*vol2;
    ordenador.sample1[3]=2*vol2;
    ordenador.sample1b[1]=1;
    ordenador.sample1b[3]=1;
    break;
  case 2:
    ordenador.sample1[0]=2*vol2;
    ordenador.sample1[2]=2*vol2;
    ordenador.sample1b[0]=1;
    ordenador.sample1b[2]=1;
    break;
  }

}
