#ifndef computer_h
#define computer_h

#include <SDL/SDL.h>

extern char salir;

struct computer {

  unsigned int temporal_io;

  // screen private global variables
  SDL_Surface *screen;
  unsigned char *screenbuffer;
  unsigned int screen_width;
  unsigned int translate[6144],translate2[6144];

  int contador_flash;

  unsigned int *p_translt,*p_translt2;
  unsigned char *pixel; // current address
  char border,flash;
  int currline,currpix;

  int tstados_counter; // counts tstates leaved to the next call
  int resx,resy,bpp; // screen resolutions
  int init_line; // cuantity to add to the base address to start to paint
  int next_line; // cuantity to add when we reach the end of line to go to next line
  int first_line; // first line to start to paint
  int last_line; // last line to paint
  int first_pixel; // first pixel of a line to paint
  int last_pixel; // last pixel of a line to paint
  int pixancho,pixalto; // maximum pixel value for width and height

  //  unsigned char videomem[112320]; // virtual video memory

  // keyboard private global variables

  unsigned char s8,s9,s10,s11,s12,s13,s14,s15;
  unsigned char k8,k9,k10,k11,k12,k13,k14,k15;
  unsigned char readed;

  // kempston joystick private global variables

  unsigned char js,jk;

  // sound global variables

  int tst_sample; // number of tstates per sample
  int freq; // frequency for reproduction
  int format; // 0: 8 bits, 1: 16 bits LSB, 2: 16 bits MSB
  signed char sign; // 0: unsigned; 1: signed
  int channels; // number of channels
  int buffer_len; // sound buffer length (in samples)
  int increment; // cuantity to add to jump to the next sample
  unsigned char volume; // volume
  unsigned char sample1[4]; // buffer with precalculated sample 1 (for buzzer)
  unsigned char sample1b[4]; // buffer with prec. sample 1 (for AY-3-8912)
  unsigned char sample0[4]; // buffer with precalculated sample 0
  unsigned char sound_bit;
  unsigned int tstados_counter_sound;
  unsigned char *current_buffer;
  unsigned char num_buff;
  unsigned int sound_cuantity; // counter for the buffer
  unsigned char ay_registers[16]; // registers for the AY emulation
  unsigned int aych_a,aych_b,aych_c,aych_n,aych_envel; // counters for AY emulation
  unsigned char ayval_a,ayval_b,ayval_c,ayval_n;
  unsigned char ay_emul; // 0: no AY emulation; 1: AY emulation
  unsigned char vol_a,vol_b,vol_c;
  unsigned int tst_ay;
  unsigned int tst_ay2;
  unsigned int ay_latch;
  char ay_envel_value;
  unsigned char ay_envel_way;

  // bus global variables

  unsigned char bus_counter;
  unsigned char bus_value;
  unsigned char issue; // 2= 48K issue 2, 3= 48K issue 3
  unsigned char mode128k; // 0=48K, 1=128K, 2=+2, 3=+3
  unsigned char joystick; // 0=cursor, 1=kempston, 2=sinclair1, 3=sinclair2
  unsigned char port254;


  // tape global variables

  unsigned char tape_mode; // 0: guide; 1: data; 2: pause; 3: not initialized
  unsigned char pause; // 1=tape stop
  unsigned int tape_counter0;
  unsigned int tape_counter1;
  unsigned int tape_counter_rep;
  unsigned char tape_byte;
  unsigned char tape_bit;
  unsigned char tape_readed;
  unsigned int tape_byte_counter;
  FILE *tap_file;
  unsigned char fast_load; // 0 normal load; 1 fast load
  unsigned char current_tap[2049];

  // OSD global variables

  unsigned char osd_text[80];
  unsigned int osd_time;

  // pagination global variables

  unsigned char mport1,mport2; // ports for memory management (128K and +3)
  unsigned int video_offset; // 0 for page 5, and 32768 for page 7
  unsigned char *block0,*block1,*block2,*block3; // pointers for memory access (one for each 16K block).

  // public

  unsigned char memoria[196608]; // memory (12 pages of 16K each one). 4 for ROM, and 8 for RAM
  unsigned char interr;
  unsigned char mustlock;
  unsigned char other_ret; // 0=no change; 1=memory returns RET (201)

};

void computer_init();
void register_screen(SDL_Surface *);
inline void show_screen(int);
inline void paint_pixels(unsigned char, char, char);
inline void read_keyboard();
void fill_audio(void *udata,Uint8 *,int);
void set_volume(unsigned char);
inline void play_sound(unsigned int);
inline void emulate(int);
void ResetComputer();
inline byte bus_empty();
inline void tape_read(FILE *,int);
void fastload_block(FILE *);
void set_memory_pointers();
inline void play_ay();
#endif
