#include "Z80.h"
#include "computer.h"
#include "emulator.h"
#include "cargador.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "characters.h"
#include "menus.h"
#include <string.h>
#include <SDL/SDL.h>


Z80 procesador;
struct computer ordenador;
SDL_Surface *screen;
char salir;
unsigned char *sound1,*sound2;
volatile struct sound_buffer snd_buffer;
SDL_mutex *semaforo;
char path_snaps[2049];
char path_taps[2049];


void load_rom(char type) {

  FILE *fichero;
  int size;

  switch(type) {
  case 0:
    fichero=fopen("./roms/spectrum.rom","r");
    size=16384;
    break;
  case 1:
    fichero=fopen("./roms/128.rom","r");
    size=32768;
    break;
  case 2:
    fichero=fopen("./roms/plus2.rom","r");
    size=32768;
    break;
  case 3:
    fichero=fopen("./roms/plus3.rom","r");
    size=65536;
    break;
  default:
    fichero=NULL;
    size=0;
  }    
  if(fichero==NULL) {
    printf("Can't open ROM file\n");
    exit(1);
  }
  fread(ordenador.memoria,size,1,fichero);
  fclose(fichero);
}

void init_screen(int resx,int resy,int depth) {

  int retorno,bucle;
  
  SDL_AudioSpec wanted,obtained;

  retorno=SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO);
  if(retorno!=0) {
    printf("Can't initialize SDL library. Exiting\n");
    exit(1);
  }

  // screen initialization

  screen=SDL_SetVideoMode(resx,resy,depth,SDL_HWSURFACE|SDL_HWPALETTE); // |SDL_ANYFORMAT); //|SDL_FULLSCREEN);
  if(screen==NULL) {
    printf("Can't assign SDL Surface. Exiting\n");
    exit(1);
  }

  if(SDL_MUSTLOCK(screen)) {
    ordenador.mustlock=1;
    SDL_LockSurface(screen);
  } else
    ordenador.mustlock=0;

  // sound initialization

  wanted.freq=48000;
  wanted.format=AUDIO_S8;
  wanted.channels=1;
  wanted.samples=1024; // 1/10 second = 350.000 TStates
  wanted.callback=fill_audio;
  wanted.userdata=(void *)&snd_buffer;

  if(SDL_OpenAudio(&wanted,&obtained)) {
    printf("Can't open audio. Exiting\n");
    exit(1);
  }

  ordenador.freq=obtained.freq;
  switch(obtained.format) {
  case AUDIO_U8:
    ordenador.sign=0;
    ordenador.format=0;
    break;
  case AUDIO_S8:
    ordenador.sign=-128;
    ordenador.format=0;
    break;
  case AUDIO_U16:
    ordenador.sign=0;
    ordenador.format=1;
    break;
  case AUDIO_S16:
    ordenador.sign=-128;
    ordenador.format=1;
    break;
  case AUDIO_U16MSB:
    ordenador.sign=0;
    ordenador.format=2;
    break;
  case AUDIO_S16MSB:
    ordenador.sign=-128;
    ordenador.format=2;
    break;
  }
  ordenador.channels=obtained.channels;
  ordenador.buffer_len=obtained.samples;
  sound1=(unsigned char *)malloc(obtained.samples+4);
  sound2=(unsigned char *)malloc(obtained.samples+4);
  for(bucle=0;bucle<obtained.samples;bucle++) {
    sound1[bucle]=0;
    sound2[bucle]=0;
  }
  snd_buffer.buffer=sound1;
  if(ordenador.format)
    ordenador.increment=2*ordenador.channels;
  else
    ordenador.increment=ordenador.channels;
  snd_buffer.length=obtained.samples/ordenador.increment;
  ordenador.tst_sample=3500000/obtained.freq;
  set_volume(70);

}

void end_system() {

  if(ordenador.mustlock)
    SDL_UnlockSurface(screen);

  if(ordenador.tap_file!=NULL)
    fclose(ordenador.tap_file);

  SDL_DestroyMutex(semaforo);
  SDL_Quit();
}


int main(int argc,char *argv[]) {

  int bucle,tstados;
  word PC;

  atexit(end_system);
  init_screen(640,480,8);

  computer_init();
  register_screen(screen);
  ordenador.interr=0;

  ordenador.screenbuffer=ordenador.screen->pixels;
  ordenador.screen_width=ordenador.screen->w;

  // assign initial values for PATH variables

  strcpy(path_snaps,"/boot/home");
  strcat(path_snaps,"/");
  strcpy(path_taps,path_snaps);
  ordenador.current_tap[0]=0;

  // assign random values to the memory before start execution

  for(bucle=0;bucle<196608;bucle++)
    ordenador.memoria[bucle]=(unsigned char) rand();

  ordenador.tap_file=NULL;

  // we filter all the events, except keyboard events

  SDL_EventState(SDL_ACTIVEEVENT,SDL_IGNORE);
  SDL_EventState(SDL_MOUSEMOTION,SDL_IGNORE);
  SDL_EventState(SDL_MOUSEBUTTONDOWN,SDL_IGNORE);
  SDL_EventState(SDL_MOUSEBUTTONUP,SDL_IGNORE);
  SDL_EventState(SDL_JOYAXISMOTION,SDL_IGNORE);
  SDL_EventState(SDL_JOYBALLMOTION,SDL_IGNORE);
  SDL_EventState(SDL_JOYHATMOTION,SDL_IGNORE);
  SDL_EventState(SDL_JOYBUTTONDOWN,SDL_IGNORE);
  SDL_EventState(SDL_JOYBUTTONUP,SDL_IGNORE);
  SDL_EventState(SDL_QUIT,SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT,SDL_IGNORE);
  SDL_EventState(SDL_VIDEORESIZE,SDL_IGNORE);
  SDL_EventState(SDL_USEREVENT,SDL_IGNORE);

  SDL_ShowCursor(SDL_DISABLE);

  salir=1;

  ResetComputer();
  
  procesador.Trace=0;
  procesador.IAutoReset=0;
  procesador.TrapBadOps=0;

  semaforo=SDL_CreateMutex();
  SDL_mutexP(semaforo); // lock it

  SDL_PauseAudio(0);

  procesador.ICount=200; // to ensure that all operations will be done
  procesador.TStates=0;

  while(salir) {

    PC=ExecZ80(&procesador);
    /* if PC is 0x0556, a call to LD_BYTES has been made, so if
       FAST_LOAD is 1, we must load the block in memory and return */

    if((PC==0x0556) && (ordenador.fast_load==1)) {
      if(ordenador.tap_file!=NULL)
	fastload_block(ordenador.tap_file);
      else {
	sprintf(ordenador.osd_text,"No TAP file selected");
	ordenador.osd_time=50;
      }
    }

    tstados=procesador.TStates;
    if(tstados==0)
      tstados=4; // if we are in halt, we execute virtual NOPs
    if(tstados<0) {
      printf("Error\n");
      exit(1);
    }
    emulate(tstados); // execute the whole hardware emulation

    procesador.ICount=200; // to ensure that all operations will be done
    procesador.TStates=0;

    if(ordenador.interr==1) {
      IntZ80(&procesador,bus_empty()); // execute the interruption
      ordenador.interr=0;
    }
  }
  return 0;
}
